{{
TwiKeyboard.spin COG device driver.

Purpose of this object is to provide communications services
between the Propeller and TWI attached keyboard.

Functions of this code:

Interface between this object and applications is over TWI via
queues. A power of 2 sized circular queue is maintained for
each device. The device driver will place data into a queue and
increment the head pointer. The user will read data and
increment the tail pointer. The user must read the queue fast
enough so that old data is not destroyed. The user will know
if data is available when head value is not equal tail.

This driver is an TWI "master" and talks to the PS2 TWI "slave".

Functions of the PS2 TWI device:

The PS2 TWI is a communications pipe. It multiplexes Keyboard
and Mouse data on the TWI interface protocol. The device can
be read and written by the master. The master writes to the
mouse address and the TWI controller passes the message on
to the mouse. Messages are similarly sent to the keyboard.
The master reads from the TWI device's input queue. The
TWI device relies on interrupts for low latency data
collection.

The key item that makes reading the PS2 device reliable is what
we call "back pressure". The controller will assert the PS2 clock
line while TWI transactions happen. Without this back pressure
feature, the design would not be usable.
}}

'=============================================================================
'
CON
{{
Normally we can use Keyboard and Mouse device IDs to talk with the PS2 device.
For the keyboard only driver, we just use one ID, but this means the keyboard
would be plugged into the MOUSE port on the atTiny85 TWI solution. The atTiny84
"Tiny Two Wire" and Propeller TWI solutions can easily handle the right ports.

The atTiny85 solution is a little handicapped by the RESET pin unless an AVR
high voltage programmer is available like the AVRDRAGON for the user.
}}

  DEVICEID      = $80           ' user's device id
  KEYBD         = DEVICEID

CON

  QSIZE         = 1 << 7        ' power of 2 queue should be at least 8 but prefer 64

  ACK           = 0             ' TWI Acknowledge
  NAK           = 1             ' TWI No Acknowledge
  RECV          = 1             ' TWI Direction Receive

  PS2RESET      = $FF           ' Common reset command byte
  PS2RESEND     = $FF           ' Common resend command byte
  PS2ACK        = $FA           ' Common acknowledge byte
  PS2DISABLE    = $F5           ' Common disable command byte
  PS2ENABLE     = $F4           ' Common enable command byte
  PS2RATE       = $F3           ' Used by mouse for ID. used by keybd for typerate
  PS2GETID      = $F2           ' Common get device ID command
  PS2MREMOTE    = $F0           ' Mouse set remote command
  PS2KSETSCAN   = $F0           ' Keyboard set scan code command
  PS2BATOK      = $AA           ' Common Basic Assurance Test OK byte
  PS2BATFAIL    = $FC           ' Common Basic Assurance Test Failure byte
  PS2NODATA     = $A5           ' Common no data available byte
  PS2SCALE221   = $E7
  PS2SCALE121   = $E6
  PS2SETLEDS    = $ED

  delay         = 10
  RETRIES       = 5             ' number of reset retries before giving up on a device

VAR     ' cog interface - must keep in this order

  long  sclmsk, sdamsk          ' pins masks for interface
  long  wrcmd,  wrdata          ' write command mailbox interface
  long  khead,  ktail           ' keybd queue pointers
  long  mhead,  mtail           ' mouse queue pointers
  long  kqptr,  mqptr           ' pointers to queues
  long  kparm,  mparm           ' parameters for keyboard and mouse


VAR

  byte kqueue[QSIZE]  ' keybd queue
  byte mqueue[QSIZE]  ' mouse queue

  long  cog           ' cog variable storage

'=============================================================================
' start the interface
'
PUB start(clkpin,datpin)
''
'' All lock-keys will be enabled, NumLock will be initially 'on',
'' and auto-repeat will be set to 15cps with a delay of .5s
'' Default keyboard startup. See startx for bitstream definitions.
''

  return startx(clkpin, datpin, %0_000_100, %01_01000)

PUB startx(clkpin, datpin, kblocks, kbrate)

'' Like start, but allows you to specify lock settings and auto-repeat
''
''   locks = lock setup
''           bit 6 disallows shift-alphas (case set soley by CapsLock)
''           bits 5..3 disallow toggle of NumLock/CapsLock/ScrollLock state
''           bits 2..0 specify initial state of NumLock/CapsLock/ScrollLock
''           (eg. %0_001_100 = disallow ScrollLock, NumLock initially 'on')
''
''   auto  = auto-repeat setup
''           bits 6..5 specify delay (0=.25s, 1=.5s, 2=.75s, 3=1s)
''           bits 4..0 specify repeat rate (0=30cps..31=2cps)
''           (eg %01_00000 = .5s delay, 30cps repeat)


  sclmsk := 1 << clkpin         ' setup pin masks
  sdamsk := 1 << datpin

  kqptr  := @kqueue             ' setup queue pointers
  mqptr  := @mqueue

  key_present := 0              ' set zero so startup works
  kparm := @par_tail
  long[kparm][2] := @keytable  ' present var has table at startup
  longmove(@par_keys,@clkpin,4) ' compatible with keyboard.spin

  wrcmd~~
  result := cog := cognew(@entry,@sclmsk)+1
  repeat while wrcmd

PUB stop
  if cog
    cogstop(cog~ - 1)

'=============================================================================
' keyboard interface
'
VAR
  long  par_tail        'key buffer tail        read/write      (19 contiguous longs)
  long  par_head        'key buffer head        read-only
  long  key_present     'keyboard present       read-only
  long  par_states[8]   'key states (256 bits)  read-only
  long  par_keys[8]     'key buffer (16 words)  read-only       (also used to pass initial parameters)

PUB getkeyhead
  return par_head
PUB getkeytail
  return par_tail

PUB kbpresent

'' Check if keyboard present - valid ~2s after start
'' returns t|f

  return key_present

PUB key : keycode

'' Get key (never waits)
'' returns key (0 if buffer empty)

  if par_tail <> par_head
    keycode := par_keys.word[par_tail]
    par_tail := ++par_tail & $F

PUB getkey : keycode

'' Get next key (may wait for keypress)
'' returns key

  repeat until (keycode := key)

PUB newkey : keycode

'' Clear buffer and get new key (always waits for keypress)
'' returns key

  par_tail := par_head
  keycode := getkey


PUB gotkey : truefalse

'' Check if any key in buffer
'' returns t|f

  truefalse := par_tail <> par_head

PUB clearkeys

'' Clear key buffer

  par_tail := par_head


PUB keystate(k) : state

'' Get the state of a particular key
'' returns t|f

  state := -(par_states[k >> 5] >> k & 1)

DAT     org
'============================================================
' TWI keyboard/mouse device COG driver
'============================================================
entry   ' get parameters then start command loop

long 0 [8]
tmp     mov     tmp,par

'------------------------------------------------------------
' initialize control/data pointers
'
_sclmsk         rdlong  $,tmp           ' scl mask
                add     tmp,#4
_sdamsk         rdlong  $,tmp           ' sda mask
                add     tmp,#4
_wrcmd          mov     $,tmp           ' _wrcmd is pointer to device id to write
                add     tmp,#4
_wrdata         mov     $,tmp           ' _wrdata is pointer to data
                add     tmp,#4
_khead          mov     $,tmp           ' pointer to khead
                add     tmp,#4
_ktail          mov     $,tmp           ' pointer to ktail
                add     tmp,#4
_mhead          mov     $,tmp           ' pointer to mhead
                add     tmp,#4
_mtail          mov     $,tmp           ' pointer to mtail
                add     tmp,#4
_kqueue         mov     $,tmp           ' pointer to kqueue address 0
                add     tmp,#4
_mqueue         mov     $,tmp           ' pointer to mqueue address 0
                add     tmp,#4
_kparm          mov     $,tmp           ' pointer to keyboard parms
                add     tmp,#4
_mparm          mov     $,tmp           ' pointer to mouse parms


'------------------------------------------------------------
' initialize keyboard pointers
'
_kptr           rdlong  $,_kparm
_kbtail         mov     $,_kptr         ' keybd tail pointer
                add     _kptr,#4
_kbhead         mov     $,_kptr         ' keybd head pointer
                add     _kptr,#4
_kbpres         mov     $,_kptr         ' pointer to keybd present
_kbtable        rdlong  $,_kptr         ' keybd present variable = keytable at startup
                add     _kptr,#4
_kbstates       mov     $,_kptr         ' keybd states
                add     _kptr,#32       ' keybd states uses 8 longs
_kbkeyqp        mov     $,_kptr         ' key buffer queue pointer

                add     _kptr,#8        ' keybd autorate parameter
                rdlong  _locks,_kptr    ' get parameter
                add     _kptr,#8        ' keybd locks parameter
                rdlong  _auto,_kptr     ' get parameter

                mov     _kbpres,#0      ' no keyboard at startup

                call    #twi_init

'------------------------------------------------------------
' permiscuous configuration protocol must pass before continuing
'
doPCP
                mov     twiDev,#DEVICEID' assign user's device id
                or      twiDev,#1       ' always a read
                call    #_write         ' send data
                call    #pdelay
        if_c    jmp     #doPCP          ' if ack ok, carry is clear

                mov     wcmd,#0         ' cog startup finished
                wrlong  wcmd,_wrcmd

'------------------------------------------------------------
' main command loop. poll device before checking for data write.
'
cmdloop
                rdlong  wcmd,_wrcmd wz
        if_nz   call    #writeDevice ' if wcmd <> 0, write wcmd device

                call    #pdelay
                call    #pollDevices    ' poll devices for data
                call    #updateKeybd

                jmp     #cmdloop
'
'------------------------------------------------------------

pdelay
                mov     ptime,#$100     ' $100 << 8 gives enough delay so next poll skips
                shl     ptime,#8        ' beginning of PS2 packet. $110 << 8 lags but works.
                add     ptime,cnt
                waitcnt ptime,ptime     ' wait a while before poll
pdelay_ret      ret

ptime   long    0

wcmd    long    0
ndx     long    0
sdata   long    0

twiAck  long    0
twiDev  long    0
twiData long    0
twiPlen long    0
twiPckt long    0

head    long    0
tail    long    0
tp      long    0
byt     long    0
butn    long    0
mtype   long    0
ktype   long    0

bext    long    $ffff_ff80
bmsk    long    $7f


dlsb    long    1 << 9

stat    long    8
data    long    0
x       long    0
y       long    0
t       long    0

_head    long    0       'write-only
_present long    0       'write-only
_states  long    0 [8]   'write-only
_locks   long    0
_auto    long    0

'------------------------------------------------------------------------------
' handle keybd data
'
'
'
' Get scancode
'
updateKeybd             'receive byte from keyboard

                        test    stat,#8         wc      'if reset flag, transmit reset command
        if_c            mov     stat,#0                 ' same functionality ask keyboad.spin
        if_c            mov     data,#PS2RESET
        if_c            call    #transmit

:same                   'call  #receive                 ' this is receive
                        rdlong  head,_khead             ' except it's cooperatively tasked
                        rdlong  tail,_ktail
                        cmp     head,tail wz            ' do we have new queue value?
        if_e            jmp     #updateKeybd_ret        ' nothing to do. exit.

                        rdlong  tp,_kqueue              ' get twi queue pointer
                        add     tp,tail                 ' get queue+tail
                        rdbyte  data,tp                 ' read byte from twi queue

                        add     tail,#1
                        and     tail,#QSIZE-1           ' roll twi rx tail
                        wrlong  tail,_ktail

                        cmp     data,#$83+1     wc      'scancode?

        if_nc           cmp     data,#$AA       wz      'powerup/reset?
        if_nc_and_z     jmp     #configure

        if_nc           cmp     data,#$E0       wz      'extended?
        if_nc_and_z     or      stat,#1
        if_nc_and_z     jmp     #updateKeybd_ret        ' back to updateKeybd is :same

        if_nc           cmp     data,#$F0       wz      'released?
        if_nc_and_z     or      stat,#2
        if_nc_and_z     jmp     #updateKeybd_ret        ' back to updateKeybd is :same

        if_nc           jmp     #newcode                'unknown, ignore. nothing to do. exit.
'
'
' Translate scancode and enter into buffer
'
                        test    stat,#1         wc      'lookup code with extended flag
                        rcl     data,#1
                        call    #look

                        cmp     data,#0         wz      'if unknown, ignore
        if_z            jmp     #newcode

                        mov     t,_states+6             'remember lock keys in _states

                        mov     x,data                  'set/clear key bit in _states
                        shr     x,#5
                        add     x,#_states
                        movd    :reg,x
                        mov     y,#1
                        shl     y,data
                        test    stat,#2         wc
:reg                    muxnc   0,y

        if_nc           cmpsub  data,#$F0       wc      'if released or shift/ctrl/alt/win, done
        if_c            jmp     #update

                        mov     y,_states+7             'get shift/ctrl/alt/win bit pairs
                        shr     y,#16

                        cmpsub  data,#$E0       wc      'translate keypad, considering numlock
        if_c            test    _locks,#%100    wz
        if_c_and_z      add     data,#@keypad1-@keytable       ' TBD check this
        if_c_and_nz     add     data,#@keypad2-@keytable       ' TBD check this
        if_c            call    #look
        if_c            jmp     #:flags

                        cmpsub  data,#$DD       wc      'handle scrlock/capslock/numlock
        if_c            mov     x,#%001_000
        if_c            shl     x,data
        if_c            andn    x,_locks
        if_c            shr     x,#3
        if_c            shr     t,#29                   'ignore auto-repeat
        if_c            andn    x,t             wz
        if_c            xor     _locks,x
        if_c            add     data,#$DD
        if_c_and_nz     or      stat,#4                 'if change, set configure flag to update leds

                        test    y,#%11          wz      'get shift into nz

        if_nz           cmp     data,#$60+1     wc      'check shift1
        if_nz_and_c     cmpsub  data,#$5B       wc
        if_nz_and_c     add     data,#@shift1-@keytable
        if_nz_and_c     call    #look
        if_nz_and_c     andn    y,#%11

        if_nz           cmp     data,#$3D+1     wc      'check shift2
        if_nz_and_c     cmpsub  data,#$27       wc
        if_nz_and_c     add     data,#@shift2-@keytable
        if_nz_and_c     call    #look
        if_nz_and_c     andn    y,#%11

                        test    _locks,#%010    wc      'check shift-alpha, considering capslock
                        muxnc   :shift,#$20
                        test    _locks,#$40     wc
        if_nz_and_nc    xor     :shift,#$20
                        cmp     data,#"z"+1     wc
        if_c            cmpsub  data,#"a"       wc
:shift  if_c            add     data,#"A"
        if_c            andn    y,#%11

:flags                  ror     data,#8                 'add shift/ctrl/alt/win flags
                        mov     x,#4                    '+$100 if shift
:loop                   test    y,#%11          wz      '+$200 if ctrl
                        shr     y,#2                    '+$400 if alt
        if_nz           or      data,#1                 '+$800 if win
                        ror     data,#1
                        djnz    x,#:loop
                        rol     data,#12

                        rdlong  x,_kbtail               'if room in buffer and key valid, enter
                        sub     x,#1
                        and     x,#$F
                        cmp     x,_head         wz
        if_nz           test    data,#$FF       wz
        if_nz           mov     x,_kbtail
        if_nz           add     x,#11*4
        if_nz           add     x,_head
        if_nz           add     x,_head
        if_nz           wrword  data,x
        if_nz           add     _head,#1
        if_nz           and     _head,#$F

                        test    stat,#4         wc      'if not configure flag, done
        if_nc           jmp     #update                 'else configure to update leds

'
'
' Configure keyboard
'
configure
                        mov     data,#$F3               'set keyboard auto-repeat
                        call    #transmit
                        mov     data,_auto
                        and     data,#%11_11111
                        call    #transmit

                        mov     data,#$ED               'set keyboard lock-leds
                        call    #transmit
                        mov     data,_locks
                        rev     data,#-3 & $1F
                        test    data,#%100      wc
                        rcl     data,#1
                        and     data,#%111
                        call    #transmit

                        mov     x,_locks                'insert locks into _states
                        and     x,#%111
                        shl     _states+7,#3
                        or      _states+7,x
                        ror     _states+7,#3

                        mov     _present,#1             'set _present
'
' update the parameters
' using original names to minimize changes in keybd driver
'
update
                        movd    :par,#_head             'update output parameters _head/_present/_states[8]
                        mov     x,_kbtail
                        add     x,#1*4
                        mov     y,#1+1+8
:par                    wrlong  0,x
                        add     :par,dlsb
                        add     x,#4
                        djnz    y,#:par
'
' newcode needs stat cleared
' using original names to minimize changes in keybd driver
'
newcode                 mov     stat,#0

updateKeybd_ret ret


'------------------------------------------------------------
' lookup keycode
' using original names to minimize changes in keybd driver
'
look            mov     tmp,_kbtable
                add     tmp,data                ' lookup keytable entry
                rdbyte  data,tmp

look_ret        ret


'------------------------------------------------------------
' send byte to keyboard
' using original names to minimize changes in keybd driver
'
transmit
                mov     twiDev,#KEYBD           ' user's device id
                andn    twiDev,#1               ' always a write
                mov     twiData,data            ' user's data
                call    #twi_wrbyte             ' send data
                call    #pdelay
transmit_ret    ret


'------------------------------------------------------------
' write device
'
writeDevice
        rdlong  sdata,_wrdata
        mov     twiDev,wcmd             ' user's device id
        call    #writeDeviceSub
        mov     ptime,#$100             ' give device a chance to respond
        shl     ptime,#11               ' before the read routine crushes clock.
        add     ptime,cnt
        waitcnt ptime,ptime             ' wait a while before poll
        mov     wcmd,#0                 ' write device finished
        wrlong  wcmd,_wrcmd
        jmp     #cmdloop
writeDevice_ret ret

writeDeviceSub
        andn    twiDev,#1               ' always a write
        mov     twiData,sdata           ' user's data
        call    #twi_wrbyte             ' send data
writeDeviceSub_ret ret

'------------------------------------------------------------
' poll devices
'
pollDevices
        call    #kbreceive
pollDevices_ret ret

devpoll long    0

kbpoll
        mov     twiDev,#KEYBD           ' check KEYBD
        mov     sdata,#$FA
        call    #writeDeviceSub
        call    #pdelay
        call    #kbreceive              ' receive the data if any
kbpoll_ret      ret

kbreceive
        mov     twiDev,#KEYBD           ' check KEYBD
        call    #twi_chkbyte
if_c    jmp     #kbreceive_ret
        or      twiAck,_sdamsk          ' just one byte
        call    #_read                  ' read byte
        call    #_stop
        call    #queueKeybd
kbreceive_ret   ret

'------------------------------------------------------------
' queue data to mouse
'
queueMouse
        rdlong  tmp,_mqueue             ' get queue start
        rdlong  ndx,_mhead              ' set position
        add     tmp,ndx
        wrbyte  twiData,tmp             ' write byte
        add     ndx,#1                  ' add to queue head
        and     ndx,#QSIZE-1            ' it's circular
        wrlong  ndx,_mhead
queueMouse_ret  ret

'------------------------------------------------------------
' queue data to keybd
'
queueKeybd
        rdlong  tmp,_kqueue             ' get queue start
        rdlong  ndx,_khead              ' set position
        add     tmp,ndx
        wrbyte  twiData,tmp             ' write byte
        add     ndx,#1                  ' add to queue head
        and     ndx,#QSIZE-1            ' it's circular
        wrlong  ndx,_khead
queueKeybd_ret  ret

'------------------------------------------------------------
' initializes twi devices to known state
' sends to address $ff which should get no ack
'
twi_init
        or      outa,_sclmsk
        or      dira,_sclmsk
        andn    dira,_sdamsk
        mov     ndx,#9
:loop   andn    outa,_sclmsk
        or      outa,_sclmsk
        test    _sdamsk,ina wc          ' test for sda high
if_nc   djnz    ndx,#:loop              ' loop if not high
twi_init_ret    ret

'------------------------------------------------------------
' chkbyte
'
' twiDev  - device id
' carry set if no data available
' is device busy ? does device have data for us ?
'
twi_chkbyte
        call    #_start
        mov     twiData,twiDev
        or      twiData,#1               ' it's a read transaction
        call    #_write                  ' get device attention
        test    twiAck,_sdamsk wc        ' if nack, return carry set
if_c    call    #_stop
twi_chkbyte_ret  ret


'------------------------------------------------------------
' wrbyte
'
' twiDev  - device id
' twiData - data to write
'
twi_wrbyte
        mov     sdata,twiData
        call    #_wait
if_nc   mov     twiData,sdata
if_nc   call    #_write
if_nc   call    #_stop
twi_wrbyte_ret  ret

'------------------------------------------------------------
' wait
'
' twiDev  - device id
' twiAck  - cleared if device is responding
' carry clear if device responding
'
_wait
        call    #_start
        call    #twi_delay
        andn    twiDev,#1               ' it's a write transaction
        mov     twiData,twiDev
        call    #_write
        test    twiAck,_sdamsk wc        ' wait for dev not busy
if_c    call    #_stop
if_c    jmp     #_wait
_wait_ret    ret

'------------------------------------------------------------
' sets start state
' start is SCL high with SDA High->Low
'
_start
        or      outa,_sclmsk            ' scl high
        or      outa,_sdamsk            ' sda high
        or      dira,_sdamsk            ' turn on sda
        or      dira,_sclmsk            ' turn on scl
        call    #twi_delay
        andn    outa,_sdamsk            ' turn off sda
        call    #twi_delay
        andn    outa,_sclmsk            ' turn off scl
        call    #_startdelay
_start_ret   ret

'------------------------------------------------------------
' sets stop state
' stop  is SCL high with SDA Low->High
'
_stop
        call    #_stopdelay
        or      outa,_sclmsk            ' scl high
        call    #twi_delay
        or      outa,_sdamsk            ' sda high
        call    #twi_delay
        andn    dira,_sclmsk            ' turn off scl
        andn    dira,_sdamsk            ' turn off sda
        call    #_stopdelay
_stop_ret    ret

'------------------------------------------------------------
' write data to twi bus
' twiData - data to write
'
_write
        call    #twi_delay
        call    #twi_delay
        ror     twiData,#8              ' use shl wc to set bit
        mov     ndx,#8
:loop
        call    #twi_delay              ' scl starts low
        shl     twiData,#1 wc           ' set c based on msb at shl
        muxc    outa,_sdamsk            ' copy c bit to outa using mask
        call    #twi_delay
        or      outa,_sclmsk            ' clock out data
        call    #twi_delay
        call    #twi_delay
        andn    outa,_sclmsk            ' scl low
        djnz    ndx,#:loop              ' next bit

        call    #twi_delay
        andn    dira,_sdamsk            ' release sda to check for ack
        call    #twi_delay
        or      outa,_sclmsk            ' scl was low - make high
        call    #twi_delay

        mov     twiAck,ina              ' get sda state
        and     twiAck,_sdamsk          ' set ack flag ... non-zero is true

        call    #twi_delay
        andn    outa,_sclmsk            ' scl ends low

        call    #twi_delay
        andn    outa,_sdamsk            ' sda ends low
        or      dira,_sdamsk            ' sda ends driven
_write_ret   ret

'------------------------------------------------------------
' read data from twi bus
' twiAck  - ack value _sdamsk
' twiData - data read
'
_read
        call    #twi_delay
        call    #twi_delay
        mov     twiData,#0              ' start with 0
        andn    dira,_sdamsk            ' sda set to input
        mov     ndx,#8
:loop
        call    #twi_delay              ' scl starts low

        or      outa,_sclmsk            ' scl goes hight to get bit
        call    #twi_delay

        shl     twiData,#1              ' addc does both muxc and shl ?
        test    _sdamsk,ina wc          ' get sda state
        muxc    twiData,#1
        call    #twi_delay

        andn    outa,_sclmsk            ' scl starts low
        call    #twi_delay
        djnz    ndx,#:loop

        or      outa,twiAck
        or      dira,_sdamsk            ' set the ack if required
        call    #twi_delay
        or      outa,_sclmsk            ' scl goes hight to get bit
        call    #twi_delay

        andn    outa,_sclmsk            ' scl ends low
        call    #twi_delay
        andn    outa,_sdamsk            ' sda ends low
_read_ret    ret

'------------------------------------------------------------
' delay at start
'
_startdelay
        mov     scnt,#5
:loop   call    #twi_delay
        djnz    scnt,#:loop
_startdelay_ret ret

'------------------------------------------------------------
' delay at stop
'
_stopdelay
        mov     scnt,#2
:loop   call    #twi_delay
        djnz    scnt,#:loop
_stopdelay_ret ret

scnt    long    0

'------------------------------------------------------------
' delay between changing bit states
'
twi_delay
        mov     dtime,dper
        add     dtime,cnt
        waitcnt dtime,dtime nr
twi_delay_ret   ret     ' 10 x 50ns = 500ns

dtime   long    0
dper    long    $100     ' bad if < $a0. $100 is conservative.
'               $3b8    ' poll per ms

        fit     $1ef
'
'============================================================

dat
'
'
' Lookup table
'                               ascii   scan    extkey  regkey  ()=keypad
'
keytable                word    $0000   '00
                        word    $00D8   '01             F9
                        word    $0000   '02
                        word    $00D4   '03             F5
                        word    $00D2   '04             F3
                        word    $00D0   '05             F1
                        word    $00D1   '06             F2
                        word    $00DB   '07             F12
                        word    $0000   '08
                        word    $00D9   '09             F10
                        word    $00D7   '0A             F8
                        word    $00D5   '0B             F6
                        word    $00D3   '0C             F4
                        word    $0009   '0D             Tab
                        word    $0060   '0E             `
                        word    $0000   '0F
                        word    $0000   '10
                        word    $F5F4   '11     Alt-R   Alt-L
                        word    $00F0   '12             Shift-L
                        word    $0000   '13
                        word    $F3F2   '14     Ctrl-R  Ctrl-L
                        word    $0071   '15             q
                        word    $0031   '16             1
                        word    $0000   '17
                        word    $0000   '18
                        word    $0000   '19
                        word    $007A   '1A             z
                        word    $0073   '1B             s
                        word    $0061   '1C             a
                        word    $0077   '1D             w
                        word    $0032   '1E             2
                        word    $F600   '1F     Win-L
                        word    $0000   '20
                        word    $0063   '21             c
                        word    $0078   '22             x
                        word    $0064   '23             d
                        word    $0065   '24             e
                        word    $0034   '25             4
                        word    $0033   '26             3
                        word    $F700   '27     Win-R
                        word    $0000   '28
                        word    $0020   '29             Space
                        word    $0076   '2A             v
                        word    $0066   '2B             f
                        word    $0074   '2C             t
                        word    $0072   '2D             r
                        word    $0035   '2E             5
                        word    $CC00   '2F     Apps
                        word    $0000   '30
                        word    $006E   '31             n
                        word    $0062   '32             b
                        word    $0068   '33             h
                        word    $0067   '34             g
                        word    $0079   '35             y
                        word    $0036   '36             6
                        word    $CD00   '37     Power
                        word    $0000   '38
                        word    $0000   '39
                        word    $006D   '3A             m
                        word    $006A   '3B             j
                        word    $0075   '3C             u
                        word    $0037   '3D             7
                        word    $0038   '3E             8
                        word    $CE00   '3F     Sleep
                        word    $0000   '40
                        word    $002C   '41             ,
                        word    $006B   '42             k
                        word    $0069   '43             i
                        word    $006F   '44             o
                        word    $0030   '45             0
                        word    $0039   '46             9
                        word    $0000   '47
                        word    $0000   '48
                        word    $002E   '49             .
                        word    $EF2F   '4A     (/)     /
                        word    $006C   '4B             l
                        word    $003B   '4C             ;
                        word    $0070   '4D             p
                        word    $002D   '4E             -
                        word    $0000   '4F
                        word    $0000   '50
                        word    $0000   '51
                        word    $0027   '52             '
                        word    $0000   '53
                        word    $005B   '54             [
                        word    $003D   '55             =
                        word    $0000   '56
                        word    $0000   '57
                        word    $00DE   '58             CapsLock
                        word    $00F1   '59             Shift-R
                        word    $EB0D   '5A     (Enter) Enter
                        word    $005D   '5B             ]
                        word    $0000   '5C
                        word    $005C   '5D             \
                        word    $CF00   '5E     WakeUp
                        word    $0000   '5F
                        word    $0000   '60
                        word    $0000   '61
                        word    $0000   '62
                        word    $0000   '63
                        word    $0000   '64
                        word    $0000   '65
                        word    $00C8   '66             BackSpace
                        word    $0000   '67
                        word    $0000   '68
                        word    $C5E1   '69     End     (1)
                        word    $0000   '6A
                        word    $C0E4   '6B     Left    (4)
                        word    $C4E7   '6C     Home    (7)
                        word    $0000   '6D
                        word    $0000   '6E
                        word    $0000   '6F
                        word    $CAE0   '70     Insert  (0)
                        word    $C9EA   '71     Delete  (.)
                        word    $C3E2   '72     Down    (2)
                        word    $00E5   '73             (5)
                        word    $C1E6   '74     Right   (6)
                        word    $C2E8   '75     Up      (8)
                        word    $00CB   '76             Esc
                        word    $00DF   '77             NumLock
                        word    $00DA   '78             F11
                        word    $00EC   '79             (+)
                        word    $C7E3   '7A     PageDn  (3)
                        word    $00ED   '7B             (-)
                        word    $DCEE   '7C     PrScr   (*)
                        word    $C6E9   '7D     PageUp  (9)
                        word    $00DD   '7E             ScrLock
                        word    $0000   '7F
                        word    $0000   '80
                        word    $0000   '81
                        word    $0000   '82
                        word    $00D6   '83             F7

keypad1                 byte    $CA, $C5, $C3, $C7, $C0, 0, $C1, $C4, $C2, $C6, $C9, $0D, "+-*/"

keypad2                 byte    "0123456789.", $0D, "+-*/"

shift1                  byte    "{|}", 0, 0, "~"

shift2                  byte    $22, 0, 0, 0, 0, "<_>?)!@#$%^&*(", 0, ":", 0, "+"
''
''
''      _________
''      Key Codes
''
''      00..DF  = keypress and keystate
''      E0..FF  = keystate only
''
''
''      09      Tab
''      0D      Enter
''      20      Space
''      21      !
''      22      "
''      23      #
''      24      $
''      25      %
''      26      &
''      27      '
''      28      (
''      29      )
''      2A      *
''      2B      +
''      2C      ,
''      2D      -
''      2E      .
''      2F      /
''      30      0..9
''      3A      :
''      3B      ;
''      3C      <
''      3D      =
''      3E      >
''      3F      ?
''      40      @
''      41..5A  A..Z
''      5B      [
''      5C      \
''      5D      ]
''      5E      ^
''      5F      _
''      60      `
''      61..7A  a..z
''      7B      {
''      7C      |
''      7D      }
''      7E      ~
''
''      80-BF   (future international character support)
''
''      C0      Left Arrow
''      C1      Right Arrow
''      C2      Up Arrow
''      C3      Down Arrow
''      C4      Home
''      C5      End
''      C6      Page Up
''      C7      Page Down
''      C8      Backspace
''      C9      Delete
''      CA      Insert
''      CB      Esc
''      CC      Apps
''      CD      Power
''      CE      Sleep
''      CF      Wakeup
''
''      D0..DB  F1..F12
''      DC      Print Screen
''      DD      Scroll Lock
''      DE      Caps Lock
''      DF      Num Lock
''
''      E0..E9  Keypad 0..9
''      EA      Keypad .
''      EB      Keypad Enter
''      EC      Keypad +
''      ED      Keypad -
''      EE      Keypad *
''      EF      Keypad /
''
''      F0      Left Shift
''      F1      Right Shift
''      F2      Left Ctrl
''      F3      Right Ctrl
''      F4      Left Alt
''      F5      Right Alt
''      F6      Left Win
''      F7      Right Win
''
''      FD      Scroll Lock State
''      FE      Caps Lock State
''      FF      Num Lock State
''
''      +100    if Shift
''      +200    if Ctrl
''      +400    if Alt
''      +800    if Win
''
''      eg. Ctrl-Alt-Delete = $6C9
''
''
'' Note: Driver will buffer up to 15 keystrokes, then ignore overflow.

CON

' *****************************************************************************
' *
' *     Terms of Use : MIT License
' *
' *****************************************************************************
'
' Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
' associated documentation files (the "Software"), to deal in the Software without restriction, including
' without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
' copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
' the following conditions:
'
' The above copyright notice and this permission notice shall be included in all copies or substantial
' portions of the Software.
'
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
' LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
' NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
' WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

